/******************************************************************************/
/***                                                                        ***/
/***  Script para criao do banco de dados utilizado como exemplo na       ***/
/***  palestra sobre LOGs em Bancos de dados, apresentada por Carlos H.     ***/
/***  Cantu no 7o. Firebird Developers Day.                                 ***/
/***                                                                        ***/
/***  Voc  livre para usar as rotinas e idias aqui apresentadas, desde   ***/
/***  que mantenha os crditos originais para o autor.                      ***/
/***                                                                        ***/
/***  Script disponvel em www.firebase.com.br                              ***/
/***                                                                        ***/
/***  Necessrio servidor Firebird 2.1 para executar esse script            ***/
/***                                                                        ***/
/******************************************************************************/

SET SQL DIALECT 3;

SET NAMES WIN1252;

CREATE DATABASE 'C:\cantu_log_fb21.FDB'
USER 'SYSDBA' PASSWORD 'masterkey'
PAGE_SIZE 8192
DEFAULT CHARACTER SET WIN1252;


/******************************************************************************/
/***                               Generators                               ***/
/******************************************************************************/

CREATE GENERATOR GEN_APAGAR_ID;
CREATE GENERATOR GEN_ARECEBER_ID;
CREATE GENERATOR GEN_LOG_DATA_ID;
CREATE GENERATOR GEN_LOG_OPERATIONS_ID;


/******************************************************************************/
/***                               Exceptions                               ***/
/******************************************************************************/


CREATE EXCEPTION LOG_EXCEPT 'Problema no log!';

SET TERM ^ ; 

CREATE PROCEDURE ADD_LOG (
    TABELA VARCHAR(31))
AS
BEGIN
  EXIT;
END^


CREATE PROCEDURE CAIXA (
    INICIO DATE,
    FIM DATE)
RETURNS (
    DATA DATE,
    HISTORICO VARCHAR(30),
    VALOR NUMERIC(9,2),
    TIPO CHAR(1))
AS
BEGIN
  SUSPEND;
END^


CREATE PROCEDURE LOG_INSERT_DATA (
    IDLOGOPER BIGINT,
    COLUMN_NAME VARCHAR(31),
    NEW_VALUE VARCHAR(2046),
    OLD_VALUE VARCHAR(2046),
    NEW_BLOB BLOB SUB_TYPE 0 SEGMENT SIZE 80,
    OLD_BLOB BLOB SUB_TYPE 0 SEGMENT SIZE 80)
AS
BEGIN
  EXIT;
END^


CREATE PROCEDURE LOG_INSERT_OPER (
    TABLE_NAME VARCHAR(31),
    OPERATION CHAR(1),
    PK1 VARCHAR(50),
    PK2 VARCHAR(50),
    PK3 VARCHAR(50))
RETURNS (
    LOGOPER_ID BIGINT)
AS
BEGIN
  SUSPEND;
END^



SET TERM ; ^

CREATE TABLE APAGAR (
    ID         INTEGER NOT NULL,
    DATA       DATE DEFAULT CURRENT_DATE NOT NULL,
    HISTORICO  VARCHAR(30),
    VALOR      NUMERIC(9,2) DEFAULT 0.00 NOT NULL
);


CREATE TABLE ARECEBER (
    ID         INTEGER NOT NULL,
    DATA       DATE DEFAULT CURRENT_DATE NOT NULL,
    HISTORICO  VARCHAR(30),
    VALOR      NUMERIC(9,2) DEFAULT 0.00 NOT NULL
);


CREATE TABLE LOG_DATA (
    ID           BIGINT NOT NULL,
    IDLOGOPER    BIGINT NOT NULL,
    COLUMN_NAME  VARCHAR(31) COLLATE WIN_PTBR,
    OLD_VALUE    VARCHAR(2046) COLLATE WIN_PTBR,
    NEW_VALUE    VARCHAR(2046) COLLATE WIN_PTBR,
    OLD_BLOB     BLOB SUB_TYPE 0 SEGMENT SIZE 80,
    NEW_BLOB     BLOB SUB_TYPE 0 SEGMENT SIZE 80
);


CREATE TABLE LOG_OPERATIONS (
    IDLOGOPER         BIGINT NOT NULL,
    TABLE_NAME        VARCHAR(31) NOT NULL COLLATE WIN_PTBR,
    OPERATION         CHAR(1) NOT NULL,
    USER_NAME         VARCHAR(64) COLLATE WIN_PTBR,
    SYSTIME           TIMESTAMP DEFAULT CURRENT_TIMESTAMP(0),
    TRANSACTIONID     INTEGER,
    CLIENT_ADDRESS    VARCHAR(255) COLLATE WIN_PTBR,
    NETWORK_PROTOCOL  VARCHAR(255) COLLATE WIN_PTBR,
    TR_ISOLATION      VARCHAR(255) COLLATE WIN_PTBR,
    PK1               VARCHAR(50) COLLATE WIN_PTBR,
    PK2               VARCHAR(50) COLLATE WIN_PTBR,
    PK3               VARCHAR(50) COLLATE WIN_PTBR
);




/******************************************************************************/
/***                              Primary Keys                              ***/
/******************************************************************************/

ALTER TABLE APAGAR ADD CONSTRAINT FK_APAGAR PRIMARY KEY (ID);
ALTER TABLE ARECEBER ADD CONSTRAINT FK_ARECEBER PRIMARY KEY (ID);
ALTER TABLE LOG_DATA ADD CONSTRAINT PK_LOG_DATA PRIMARY KEY (ID);
ALTER TABLE LOG_OPERATIONS ADD CONSTRAINT PK_LOG_OPERATIONS PRIMARY KEY (IDLOGOPER);


/******************************************************************************/
/***                              Foreign Keys                              ***/
/******************************************************************************/

ALTER TABLE LOG_DATA ADD CONSTRAINT FK_LOG_DATA_IDOPER FOREIGN KEY (IDLOGOPER) REFERENCES LOG_OPERATIONS (IDLOGOPER) ON DELETE CASCADE ON UPDATE CASCADE;


/******************************************************************************/
/***                                Indices                                 ***/
/******************************************************************************/

CREATE INDEX APAGAR_DATA ON APAGAR (DATA);
CREATE INDEX ARECEBER_DATA ON ARECEBER (DATA);
CREATE INDEX IDX_LOG_DATA_COLUNA_IDOPER ON LOG_DATA (COLUMN_NAME, IDLOGOPER);
CREATE INDEX LOG_OPERATIONS_DATAHORA ON LOG_OPERATIONS (SYSTIME);
CREATE INDEX LOG_OPERATIONS_TABELA ON LOG_OPERATIONS (TABLE_NAME);


/******************************************************************************/
/***                                Triggers                                ***/
/******************************************************************************/


SET TERM ^ ;



/******************************************************************************/
/***                          Triggers for tables                           ***/
/******************************************************************************/



/* Trigger: APAGAR_BI_BI */
CREATE TRIGGER APAGAR_BI_BI FOR APAGAR
ACTIVE BEFORE INSERT POSITION 0
AS
BEGIN
  IF (NEW.ID IS NULL) THEN
    NEW.ID = GEN_ID(GEN_APAGAR_ID,1);
END
^


/* Trigger: ARECEBER_BI_BI */
CREATE TRIGGER ARECEBER_BI_BI FOR ARECEBER
ACTIVE BEFORE INSERT POSITION 0
AS
BEGIN
  IF (NEW.ID IS NULL) THEN
    NEW.ID = GEN_ID(GEN_ARECEBER_ID,1);
END
^


/* Trigger: LOG_DATA_BI */
CREATE TRIGGER LOG_DATA_BI FOR LOG_DATA
ACTIVE BEFORE INSERT POSITION 0
AS
BEGIN
  IF (NEW.ID IS NULL) THEN
    NEW.ID = GEN_ID(GEN_LOG_DATA_ID,1);
END
^


/* Trigger: LOG_DATA_BUD0 */
CREATE TRIGGER LOG_DATA_BUD0 FOR LOG_DATA
ACTIVE BEFORE UPDATE OR DELETE POSITION 0
AS
begin
  exception log_except 'Impossvel editar/apagar dados do LOG!';
end
^


/* Trigger: LOG_OPERATIONS_BI */
CREATE TRIGGER LOG_OPERATIONS_BI FOR LOG_OPERATIONS
ACTIVE BEFORE INSERT POSITION 0
AS
BEGIN
  IF (NEW.IDLOGOPER IS NULL) THEN
    NEW.IDLOGOPER = GEN_ID(GEN_LOG_OPERATIONS_ID,1);
END
^


/* Trigger: LOG_OPERATIONS_BUD0 */
CREATE TRIGGER LOG_OPERATIONS_BUD0 FOR LOG_OPERATIONS
ACTIVE BEFORE UPDATE OR DELETE POSITION 0
AS
begin
  exception log_except 'Impossvel editar dados do LOG!';
end
^


SET TERM ; ^



/******************************************************************************/
/***                           Stored Procedures                            ***/
/******************************************************************************/


SET TERM ^ ;

ALTER PROCEDURE ADD_LOG (
    TABELA VARCHAR(31))
AS
declare variable XSQL varchar(32765);
declare variable FNAME varchar(32);
declare variable FTYPE smallint;
declare variable TMPSTR varchar(255);
declare variable XDDL blob sub_type 1 segment size 256;
declare variable TAMANHO integer;
declare variable PK1 varchar(31);
declare variable PK2 varchar(31);
declare variable PK3 varchar(31);
declare variable PKTEMP varchar(31);
begin
  TABELA = upper(TABELA);

  if ((:TABELA = 'LOG_DATA') or (:TABELA = 'LOG_OPERATIONS')) then
    exception LOG_EXCEPT 'Nao  possvel criar logs nas prprias tabelas de logs!';

  XSQL = '';
  for select upper(trim(F.RDB$FIELD_NAME)), FS.RDB$FIELD_TYPE,
             FS.RDB$FIELD_LENGTH
      from RDB$RELATION_FIELDS F
      left join RDB$FIELDS FS on FS.RDB$FIELD_NAME = F.RDB$FIELD_SOURCE
      where F.RDB$RELATION_NAME = :TABELA
      order by F.RDB$FIELD_POSITION
      into :FNAME, FTYPE, :TAMANHO
  do
  begin
    XSQL = XSQL || 'if (new.' || :FNAME || ' is distinct from old.' || :FNAME ||
           ') then' || ascii_char(13);
    if (:FTYPE = 261) then -- blob
      XSQL = XSQL || ' execute procedure log_insert_data (:IDOper,''' || :FNAME ||
             ''', null, null, new.' || :FNAME || ', old.' || :FNAME || ');' || ascii_char(13);
    else
      if ((:FTYPE in (14, 37)) and
          (:TAMANHO > 2046)) then -- VARCHAR ou CHAR maior que 2046
        XSQL = XSQL || ' execute procedure log_insert_data (:IDOper,''' || :FNAME ||
               ''', substring(new.' || :FNAME || ' from 1 for 2046), substring(old.' ||
               :FNAME || ' from 1 for 2046), null, null);' || ascii_char(13);
      else
        XSQL = XSQL || ' execute procedure log_insert_data (:IDOper,''' || :FNAME ||
               ''', new.' || :FNAME || ', old.' || :FNAME || ', null, null);' || ascii_char(13);
  end

  TMPSTR = 'if (Inserting) then Oper = ''I'';' || ascii_char(13) || 'else if (Updating) then Oper = ''U'';' || ascii_char(13) || ' else if (Deleting) then Oper = ''D'';';

  /* Recupera a sequencia de chaves primarias da tabela */

  PK1 = null;
  PK2 = null;
  PK3 = null;
  for select trim(I.RDB$FIELD_NAME)
      from RDB$RELATION_CONSTRAINTS RC
      join RDB$INDEX_SEGMENTS I on (I.RDB$INDEX_NAME = RC.RDB$INDEX_NAME)
      join RDB$INDICES IDX on (IDX.RDB$INDEX_NAME = RC.RDB$INDEX_NAME)
      where (RC.RDB$CONSTRAINT_TYPE = 'PRIMARY KEY') and
            (RC.RDB$RELATION_NAME = :TABELA)
      order by I.RDB$FIELD_POSITION
      into :PKTEMP
  do
  begin
    if (PK1 is null) then
      PK1 = PKTEMP;
    else
      if (PK2 is null) then
        PK2 = PKTEMP;
      else
        if (PK3 is null) then
          PK3 = PKTEMP;
        else
          exception LOG_EXCEPT 'Chave primaria composta por mais de 3 campos!';
  end

  XDDL = 'CREATE OR ALTER TRIGGER CHCLOG_' || :TABELA || ' for ' || :TABELA ||
  ' active after insert or update or delete position 32767' || ascii_char(13) || 'AS' ||
  ascii_char(13) || 'declare variable Oper char(1) = null;' || ascii_char(13) ||
  'declare variable idoper BigInt;' || ascii_char(13) || 'begin' || ascii_char(13) || TMPSTR ||
  'if (:Oper is null) then exit;' || ascii_char(13) || 'if (:Oper = ''I'') then ' ||
  ascii_char(13) || ' execute procedure log_insert_oper (''' || :TABELA || ''',:Oper,' ||
  coalesce('new.' || :PK1, 'null') || ',' || coalesce('new.' || :PK2, 'null') || ',' ||
  coalesce('new.' || :PK3, 'null') || ') returning_values :IDOper;' || ascii_char(13) || 'else' ||
  ascii_char(13) || ' execute procedure log_insert_oper (''' || :TABELA || ''',:Oper,' ||
  coalesce('old.' || :PK1, 'null') || ',' || coalesce('old.' || :PK2, 'null') || ',' ||
  coalesce('old.' || :PK3, 'null') || ') returning_values :IDOper;' || ascii_char(13) || XSQL || 'end';

  if (XSQL <> '') then
    execute statement :XDDL;
end^


ALTER PROCEDURE CAIXA (
    INICIO DATE,
    FIM DATE)
RETURNS (
    DATA DATE,
    HISTORICO VARCHAR(30),
    VALOR NUMERIC(9,2),
    TIPO CHAR(1))
AS
DECLARE VARIABLE SALDO_ANT_CREDITO NUMERIC(9,2);
DECLARE VARIABLE SALDO_ANT_DEBITO NUMERIC(9,2);
DECLARE VARIABLE SALDO_FINAL NUMERIC(9,2);
begin

  /* Primeiro recuperamos a soma dos valores dos debitos anteriores ao
     periodo informado para posteriormente calcularmos nosso saldo inicial */

  select sum(ap.valor)
  from apagar ap
  where (ap.data < :inicio)
  into :saldo_ant_debito;

  /* Agora recuperamos a soma dos valores dos creditos anteriores ao
     periodo informado  */

  select sum(ar.valor)
  from areceber ar
  where (ar.data < :inicio)
  into :saldo_ant_credito;

  /* Caso no haja lancamentos anteriores, a variavel ficara no estado NULL,
     como no SQL qualquer operao matemtica feita com valores NULL sempre
     resulta em NULL, temos que verificar isso e atribuir 0.00 na variavel nula
     para que fornea o resultado esperado */

  if (:saldo_ant_credito is null) then saldo_ant_credito = 0.00;
  if (:saldo_ant_debito is null) then saldo_ant_debito = 0.00;

  data = :inicio - 1;
  historico = 'SALDO ANTERIOR';
  valor = (saldo_ant_credito - saldo_ant_debito);

  /* Inicializando a variavel que armazenara o saldo final
     com o valor do saldo anterior */
  saldo_final = valor;

  /* Por questes "estticas" no queremos ter valores negativos no campo valor,
     sendo assim, se o saldo anterior  negativo, devemos informar que o lanamento 
     de dbito e tornar o valor do saldo positivo, caso contratio informamos que o
     lanamento  de crdito.*/

  if (:valor < 0) then
  begin
    tipo = 'D';
    valor = valor * -1;
  end
  else
  begin
    tipo = 'C';
  end

  suspend; /* Retornando a primeira linha (registro) com o saldo anterior */

  /* o comando a seguir seleciona os registros dentro do periodo informado
     buscando na tabela de contas a pagar */

  tipo = 'D'; /* Estamos recuperando dbitos */

  for select ap.data, ap.historico, ap.valor
      from apagar ap
      where (ap.data between :inicio and :fim)
      into :data, :historico, :valor
  do
  begin
    /* Subtraimos o valor do lanamento do saldo final */
    saldo_final = saldo_final - valor;
    /* O suspend a seguir retorna uma "linha" contendo os campos
       data, historico, tipo e valor que tem seus valores armazenados
       nas variveis de mesmo nome */
    suspend;
  end

  tipo = 'C'; /* Estamos recuperando crditos */

  /* seleciona os registros dentro do periodo informado
     buscando na tabela de contas a receber */
  for select ar.data, ar.historico, ar.valor
      from areceber ar
      where (ar.data between :inicio and :fim)
      into :data, :historico, :valor
  do
  begin
    /* Adicionamos o valor do lanamento ao saldo final */
    saldo_final = saldo_final + valor;
    /* O suspend a seguir retorna uma "linha" contendo os campos
       data, historico, tipo e valor que tem seus valores armazenados
       nas variveis de mesmo nome */
    suspend;
  end

  /* Montamos o ultimo "registro" que conter os valores do saldo FINAL */
  data = :fim;
  historico = 'SALDO FINAL';
  tipo = 'C';
  valor = saldo_final;

  suspend;

end^


ALTER PROCEDURE LOG_INSERT_DATA (
    IDLOGOPER BIGINT,
    COLUMN_NAME VARCHAR(31),
    NEW_VALUE VARCHAR(2046),
    OLD_VALUE VARCHAR(2046),
    NEW_BLOB BLOB SUB_TYPE 0 SEGMENT SIZE 80,
    OLD_BLOB BLOB SUB_TYPE 0 SEGMENT SIZE 80)
AS
begin
 insert into log_data (id, idlogoper, column_name, old_value, new_value, old_blob, new_blob)
 values (gen_id(gen_log_data_id,1), :idlogoper, :column_name, :old_value, :new_value, :old_blob, :new_blob);
end^


ALTER PROCEDURE LOG_INSERT_OPER (
    TABLE_NAME VARCHAR(31),
    OPERATION CHAR(1),
    PK1 VARCHAR(50),
    PK2 VARCHAR(50),
    PK3 VARCHAR(50))
RETURNS (
    LOGOPER_ID BIGINT)
AS
declare variable S varchar(16384);
begin
   insert into log_operations (table_name, operation, user_name, systime, transactionid, client_address, network_protocol, tr_isolation, PK1, PK2, PK3)
   values (:table_name, :operation, current_user, CURRENT_TIMESTAMP(0),
          current_transaction, rdb$get_context('SYSTEM', 'CLIENT_ADDRESS'), rdb$get_context('SYSTEM', 'NETWORK_PROTOCOL'),
          rdb$get_context('SYSTEM', 'ISOLATION_LEVEL'), :PK1, :PK2, :PK3)
   returning IDLOGOPER into :logoper_id;
end^

SET TERM ; ^

commit;
