/******************************************************************************/
/***                                                                        ***/
/***  Script para criao do banco de dados utilizado como exemplo na       ***/
/***  palestra sobre LOGs em Bancos de dados, apresentada por Carlos H.     ***/
/***  Cantu no 7o. Firebird Developers Day.                                 ***/
/***                                                                        ***/
/***  Voc  livre para usar as rotinas e idias aqui apresentadas, desde   ***/
/***  que mantenha os crditos originais para o autor.                      ***/
/***                                                                        ***/
/***  Script disponvel em www.firebase.com.br                              ***/
/***                                                                        ***/
/***  Necessrio servidor Firebird 2.5 para gerao dos logs em BD          ***/
/***  externo.                                                              ***/
/***                                                                        ***/
/******************************************************************************/

SET SQL DIALECT 3;

SET NAMES WIN1252;

CREATE DATABASE 'C:LOG_25.FDB'
USER 'SYSDBA' PASSWORD 'masterkey'
PAGE_SIZE 8192
DEFAULT CHARACTER SET WIN1252;



/******************************************************************************/
/***                               Generators                               ***/
/******************************************************************************/

CREATE GENERATOR GEN_LOG_DATA_ID;
CREATE GENERATOR GEN_LOG_OPERATIONS_ID;


/******************************************************************************/
/***                               Exceptions                               ***/
/******************************************************************************/

CREATE EXCEPTION LOG_EXCEPT 'Problema no log!';



/******************************************************************************/
/***                                 Tables                                 ***/
/******************************************************************************/



CREATE TABLE LOG_DATA (
    ID           BIGINT NOT NULL,
    IDLOGOPER    BIGINT NOT NULL,
    COLUMN_NAME  VARCHAR(31) COLLATE WIN_PTBR,
    OLD_VALUE    VARCHAR(2046) COLLATE WIN_PTBR,
    NEW_VALUE    VARCHAR(2046) COLLATE WIN_PTBR,
    OLD_BLOB     BLOB SUB_TYPE 0 SEGMENT SIZE 80,
    NEW_BLOB     BLOB SUB_TYPE 0 SEGMENT SIZE 80
);


CREATE TABLE LOG_OPERATIONS (
    IDLOGOPER         BIGINT NOT NULL,
    TABLE_NAME        VARCHAR(31) NOT NULL COLLATE WIN_PTBR,
    OPERATION         CHAR(1) NOT NULL,
    USER_NAME         VARCHAR(64) COLLATE WIN_PTBR,
    SYSTIME           TIMESTAMP DEFAULT CURRENT_TIMESTAMP(0),
    TRANSACTIONID     INTEGER,
    CLIENT_ADDRESS    VARCHAR(255) COLLATE WIN_PTBR,
    NETWORK_PROTOCOL  VARCHAR(255) COLLATE WIN_PTBR,
    TR_ISOLATION      VARCHAR(255) COLLATE WIN_PTBR,
    PK1               VARCHAR(50) COLLATE WIN_PTBR,
    PK2               VARCHAR(50) COLLATE WIN_PTBR,
    PK3               VARCHAR(50) COLLATE WIN_PTBR
);




/******************************************************************************/
/***                              Primary Keys                              ***/
/******************************************************************************/

ALTER TABLE LOG_DATA ADD CONSTRAINT PK_LOG_DATA PRIMARY KEY (ID);
ALTER TABLE LOG_OPERATIONS ADD CONSTRAINT PK_LOG_OPERATIONS PRIMARY KEY (IDLOGOPER);


/******************************************************************************/
/***                              Foreign Keys                              ***/
/******************************************************************************/

ALTER TABLE LOG_DATA ADD CONSTRAINT FK_LOG_DATA_IDOPER FOREIGN KEY (IDLOGOPER) REFERENCES LOG_OPERATIONS (IDLOGOPER) ON DELETE CASCADE ON UPDATE CASCADE;


/******************************************************************************/
/***                                Indices                                 ***/
/******************************************************************************/

CREATE INDEX IDX_LOG_DATA_COLUNA_IDOPER ON LOG_DATA (COLUMN_NAME, IDLOGOPER);
CREATE INDEX LOG_OPERATIONS_DATAHORA ON LOG_OPERATIONS (SYSTIME);
CREATE INDEX LOG_OPERATIONS_TABELA ON LOG_OPERATIONS (TABLE_NAME);


/******************************************************************************/
/***                                Triggers                                ***/
/******************************************************************************/


SET TERM ^ ;



/******************************************************************************/
/***                          Triggers for tables                           ***/
/******************************************************************************/



/* Trigger: LOG_DATA_BI */
CREATE TRIGGER LOG_DATA_BI FOR LOG_DATA
ACTIVE BEFORE INSERT POSITION 0
AS
BEGIN
  IF (NEW.ID IS NULL) THEN
    NEW.ID = GEN_ID(GEN_LOG_DATA_ID,1);
END
^


/* Trigger: LOG_DATA_BUD0 */
CREATE TRIGGER LOG_DATA_BUD0 FOR LOG_DATA
ACTIVE BEFORE UPDATE OR DELETE POSITION 0
AS
begin
  exception log_except 'Impossvel editar/apagar dados do LOG!';
end
^


/* Trigger: LOG_OPERATIONS_BI */
CREATE TRIGGER LOG_OPERATIONS_BI FOR LOG_OPERATIONS
ACTIVE BEFORE INSERT POSITION 0
AS
BEGIN
  IF (NEW.IDLOGOPER IS NULL) THEN
    NEW.IDLOGOPER = GEN_ID(GEN_LOG_OPERATIONS_ID,1);
END
^


/* Trigger: LOG_OPERATIONS_BUD0 */
CREATE TRIGGER LOG_OPERATIONS_BUD0 FOR LOG_OPERATIONS
ACTIVE BEFORE UPDATE OR DELETE POSITION 0
AS
begin
  exception log_except 'Impossvel editar dados do LOG!';
end
^


SET TERM ; ^

